//interfacemultiplayer.cpp
//the multiplayer game setup menus

#include <OpenGL/gl.h>
#include "gametime.h"
#include <string.h>
#include <stdio.h>
#include "controls.h"
#include "text.h"
#include "interface.h"
#include "screen.h"
#include "network.h"
#include "gameinitexit.h"
#include "gamesystem.h"
#include "carphysics.h"
#include "carselection.h"
#include "mapselection.h"
#include "config.h"
#include "interfaceutil.h"
#include "tracker.h"
#include "random.h"
#include "gamesound.h"
#include "error.h"
#include "initexit.h"
#include "environment.h"
#include "localtracker.h"
#include "gamemem.h"
#include "writeout.h"

enum{
	kMultiplayerChat,
	kMultiplayerSelectCar,
	kMultiplayerSelectMap,
	kMultiplayerGameMode,
	kMultiplayerHostControls,
	//kMultiplayerMiniGame,
	kMultiplayerStartGame,
	kMultiplayerExit,
	kNumMultiplayerOptions,
	kMultiplayerStartGameSignal,
	kMultiplayerFailed,
	kMultiplayerKicked
};

typedef struct{
	float timeStamp;
	UInt8 playerID;
} tPingReply;

enum{
	kChatFlagAlert=1<<0,
	kChatFlagSystem=1<<1
};

#define kChatBufferMaxSize 256
#define kChatBufferMaxLines 8

typedef struct{
	tChatMessage chatBuffer[kChatBufferMaxLines];
	int chatBufferLines;
}tChatBuffer;

typedef struct{
	float t;
	tInterfaceMenuDescribtion *menu;
	tGameInfo *gInfo;
	float lastSendTime;
	int playerCar,playerID;
	int locked;
	UInt8 playerColor;
	tChatBuffer cb;
} tMultiplayerMenuData;

int gConfirmedVersion=-1;

#define S32ChangeToNative(value) ((value)=EndianS32_BtoN(value))
#define U32ChangeToNative(value) ((value)=EndianU32_BtoN(value))
#define S32ChangeToNet(value) ((value)=EndianS32_NtoB(value))
#define U32ChangeToNet(value) ((value)=EndianU32_NtoB(value))

void AddNewPlayer(tGameInfo *gInfo,int netID)
{	
	//answer with a message telling the player his ID
	NetworkSendPacket(kMessageTypeID,&(gInfo->numNetPlayers),sizeof(UInt8),kMessagePriorityHigh,netID);
	//and update the gInfo structure 
	if(gInfo->numNetPlayers)
		gInfo->playerCars[gInfo->numNetPlayers]=-1;
	gInfo->netID[gInfo->numNetPlayers]=netID;
	gInfo->playerInited[gInfo->numNetPlayers]=false;
	gInfo->playerVersions[gInfo->numNetPlayers]=0;
	gInfo->numNetPlayers++;
	if(gInfo->numPlayers<gInfo->numNetPlayers)
		gInfo->numPlayers=gInfo->numNetPlayers;

	gInfo->version++;
	gInfo->playerVersions[0]=gInfo->version;
}

void FixGInfoForLeftPlayers(tGameInfo *gInfo)
{
	int minPlayerLeft=-1;
	int playerLeft;
	PrintConsoleString("Fixing GInfo");
	do
	{
		playerLeft=-1;
		
		for(int i=1;i<gInfo->numNetPlayers;i++)
			if(gInfo->netID[i]==-1)
			{
				playerLeft=i;		
				minPlayerLeft=i;
			}
				
		if(playerLeft!=-1)
		{
			PrintConsoleString("Player %d left",playerLeft);
			//shift down all the player information for players with higher player ids
			for(UInt8 i=playerLeft;i<gInfo->numPlayers-1;i++)
			{
				gInfo->playerCars[i]=gInfo->playerCars[i+1];
				gInfo->netID[i]=gInfo->netID[i+1];
				gInfo->ping[i]=gInfo->ping[i+1];
				gInfo->playerAddOns[i]=gInfo->playerAddOns[i+1];
				gInfo->playerColors[i]=gInfo->playerColors[i+1];
				gInfo->playerVersions[i]=gInfo->playerVersions[i+1];
				gInfo->playerInited[i]=gInfo->playerInited[i+1];
				strcpy(gInfo->playerCarNames[i],gInfo->playerCarNames[i+1]);
				strcpy(gInfo->playerNames[i],gInfo->playerNames[i+1]);

			}
			for(UInt8 i=gInfo->numPlayers-1;i<kMaxPlayers;i++)
			{
				gInfo->playerCars[i]=-1;
				gInfo->netID[i]=-1;
				gInfo->playerCarNames[i][0]='\0';
				gInfo->playerNames[i][0]='\0';
			}
			gInfo->numPlayers--;
			gInfo->numNetPlayers--;
		}
	}
	while(playerLeft!=-1);
	
	if(minPlayerLeft!=-1)	
		for(UInt8 i=minPlayerLeft;i<gInfo->numNetPlayers;i++)
			//send a message to the player telling him/her to update his/her info
		{
			NetworkSendPacket(kMessageTypeID,&i,sizeof(UInt8),kMessagePriorityHigh,gInfo->netID[i]);		
			PrintConsoleString("member %d is now player %d",gInfo->netID[i],i);
		}
	gInfo->version++;
	gInfo->playerVersions[0]=gInfo->version;
}

//processes network packets while the host is waiting for players to join
void WaitForPlayers(tGameInfo *gInfo,tChatMessage *chatReceive)
{
	chatReceive->str[0]='\0';
	float timeStamp=TimeGetSeconds();
	
	//received any network messages?
	tNetworkPacket message;
	if(NetworkReceivePacket(&message))
	{
		switch(message.what)
		{		
			//has a player joined?
			case kMessageTypePlayerJoined:
				{
					UInt32 reply=kVersion;
					U32ChangeToNet(reply);
					NetworkSendPacket(kMessageTypeVersion,&reply,sizeof(UInt32),kMessagePriorityHigh,message.from);				
					PrintConsoleString("Player %d joined, NT#:%d",gInfo->numNetPlayers,message.from);
				}
				break;
				
			case kMessageTypeVersion:	
				U32ChangeToNative(*(UInt32*)message.data);
				if(*((UInt32*)message.data)==kVersion)			
					AddNewPlayer(gInfo,message.from);
				else
				{
					UInt8 reason=kJoinDeniedVersion;
					NetworkSendPacket(kMessageTypeJoinDenied,&reason,sizeof(UInt8),kMessagePriorityHigh,message.from);									
				}
				break;
				
			//a player has left
			case kMessageTypePlayerLeft:
				{
					int netID=message.from;
					int id=-1;
					//find the players id
					for(int i=0;i<gInfo->numNetPlayers;i++)
						if(gInfo->netID[i]==netID)	
							id=i;
					
					if(id!=-1)
					{
						tChatMessage m;
						m.flags=kChatFlagSystem;
						sprintf(m.str,"### Player %s left the game.",gInfo->playerNames[id]);
						PlayInterfaceSound(FileGetReference("chat.wav"));
						NetworkSendPacket(kMessageTypeChat,&m,sizeof(m),kMessagePriorityHigh,kMessageSendToAll);					

						gInfo->netID[id]=-1;
						FixGInfoForLeftPlayers(gInfo);
					}
				}
				break;
			
			//a player has told us which car he wishes to use:
			case kMessageTypeCarID:
				//update the gInfo structure 
				if(_stricmp(gInfo->playerNames[((tCarIDMessage*)message.data)->playerID],((tCarIDMessage*)message.data)->playerName))
				{
					strcpy(gInfo->playerNames[((tCarIDMessage*)message.data)->playerID],((tCarIDMessage*)message.data)->playerName);
					tChatMessage m;
					m.flags=kChatFlagSystem;
					PlayInterfaceSound(FileGetReference("chat.wav"));
					sprintf(m.str,"### Player %s joined the game.",gInfo->playerNames[((tCarIDMessage*)message.data)->playerID]);
					NetworkSendPacket(kMessageTypeChat,&m,sizeof(m),kMessagePriorityHigh,kMessageSendToAll);					
				}
				
				if(gInfo->playerCars[((tCarIDMessage*)message.data)->playerID]!=FileGetReference(((tCarIDMessage*)message.data)->carName))
				{
					gInfo->playerCars[((tCarIDMessage*)message.data)->playerID]=FileGetReference(((tCarIDMessage*)message.data)->carName);
				//	gInfo->checksums[((tCarIDMessage*)message.data)->playerID]=((tCarIDMessage*)message.data)->checksum;
				//	U32ChangeToNative(gInfo->checksums[((tCarIDMessage*)message.data)->playerID]);
					
					gInfo->playerInited[((tCarIDMessage*)message.data)->playerID]=true;
					strcpy(gInfo->playerCarNames[((tCarIDMessage*)message.data)->playerID],((tCarIDMessage*)message.data)->carName);
					gInfo->version++;
					gInfo->playerVersions[0]=gInfo->version;
				}
				gInfo->playerColors[((tCarIDMessage*)message.data)->playerID]=((tCarIDMessage*)message.data)->color;
				break;
			
			//we have received a chat message
			case kMessageTypeChat:
				//fill in the chat receive buffer
				*chatReceive=*(tChatMessage*)message.data;
				if(message.from!=0)
					if(((tChatMessage*)message.data)->flags&kChatFlagAlert)
					{
						PlayInterfaceSound(FileGetReference("chat.wav"));
						SystemNotify();
					}
					else
						PlayInterfaceSound(FileGetReference("chat.wav"));
				break;
			
			//we have received a Ping reply from a player
			case kMessageTypePong:
				//calculate players response time and update gInfo structure
				gInfo->ping[((tPingReply*)message.data)->playerID]=timeStamp-((tPingReply*)message.data)->timeStamp;			
				break;			

			case kMessageTypeVersionConfirmed:
				S32ChangeToNative(((int*)message.data)[0]);	
				S32ChangeToNative(((int*)message.data)[1]);	
				if(gInfo->playerVersions[((int*)message.data)[1]]<((int*)message.data)[0])
					gInfo->playerVersions[((int*)message.data)[1]]=((int*)message.data)[0];
				break;		
					
			case kMessageTypeVersionDenied:
				S32ChangeToNative(((int*)message.data)[1]);	
				gInfo->playerVersions[((int*)message.data)[1]]=-1;
				break;			
		}
		NetworkDisposePacket(&message);
	}
}

void ConvertGInfoSwap(tGameInfo *gInfo)
{
	S32ChangeToNative(*((SInt32*)(&(gInfo->maxTime))));
	S32ChangeToNative(*((SInt32*)(&(gInfo->silverTime))));
	S32ChangeToNative(*((SInt32*)(&(gInfo->goldTime))));
	U32ChangeToNative(gInfo->version);
	for(int i=0;i<kMaxPlayers;i++)
	{
		S32ChangeToNative(gInfo->playerAddOns[i]);
		S32ChangeToNative(gInfo->playerVersions[i]);		
		S32ChangeToNative(*((SInt32*)(&(gInfo->ping[i]))));
	}
}

//Send out a packet telling players to start the game
void StartNetGame(tGameInfo *gInfo)
{
	ConvertGInfoSwap(gInfo);
	NetworkSendPacket(kMessageTypeStart,gInfo,sizeof(tGameInfo),kMessagePriorityHigh,kMessageSendToAllButSelf);
	ConvertGInfoSwap(gInfo);
}

//Send out Ping requests to all the clients
void GatherPingStats(tGameInfo *gInfo)
{
	for(int i=0;i<gInfo->numNetPlayers;i++)
		gInfo->ping[i]=NetworkGetPlayerPing(gInfo->netID[i]);
}

//processes network packets while the client is waiting for the host to give the start signal
//returns true if start signal is recieved, false otherwise.
int WaitForStartSignal(tGameInfo *gInfo,tChatMessage *chatReceive,int *id,int *exit)
{
	chatReceive->str[0]='\0';
	//received any network messages?
	tNetworkPacket message;
	if(NetworkReceivePacket(&message))
	{
		switch(message.what)
		{
			//got the start signal?
			case kMessageTypeStart:		
				*gInfo=*(tGameInfo*)message.data;
				ConvertGInfoSwap(gInfo);
				NetworkDisposePacket(&message);
				*exit=kMultiplayerStartGameSignal;
				return true;
			
			//got an GameInfo structure update?
			case kMessageTypeGameInfo:
				*gInfo=*(tGameInfo*)message.data;
				ConvertGInfoSwap(gInfo);
				//GatherPingStats(gInfo);
				break;
			
			//got a chat message?
			case kMessageTypeChat:
				//fill in the chat receive buffer
				*chatReceive=*(tChatMessage*)message.data;
				if(message.from!=gInfo->netID[*id])
					if(((tChatMessage*)message.data)->flags&kChatFlagAlert)
					{
						PlayInterfaceSound(FileGetReference("chat.wav"));
						SystemNotify();
					}
					else
						PlayInterfaceSound(FileGetReference("chat.wav"));
				break;
			
			//got a Ping request?
			case kMessageTypePing:
				//send out a reply packet with the original ping packets time stamp and our player ID.
				tPingReply pong;
				pong.timeStamp=*(float*)message.data;
				pong.playerID=*id;
				NetworkSendPacket(kMessageTypePong,&pong,sizeof(tPingReply),kMessagePriorityLow,kMessageSendToHostOnly);
				break;

			//change of our id
			case kMessageTypeID:
				*id=*(UInt8*)message.data;
				PrintConsoleString("we are now player %d",*id);
				break;
			
			//game terminated by host
			case kMessageTypeGameTerminated:
				*exit=kMultiplayerFailed;
				NetworkDisposePacket(&message);
				return true;

			case kMessageTypeKicked:
				*exit=kMultiplayerKicked;
				NetworkDisposePacket(&message);
				return true;
			
		}
		NetworkDisposePacket(&message);
	}
	return false;
}

//tell the other players on the network that player has chosen a new car
void UpdatePlayerCar(tMultiplayerMenuData *userData)
{
	tCarIDMessage message;
	message.playerID=userData->playerID;
	message.color=userData->playerColor;
//	message.checksum=FileGetChecksum(userData->playerCar);
//	U32ChangeToNet(message.checksum);
	strcpy(message.playerName,gConfig->playerName);
	strcpy(message.carName,FileGetName(userData->playerCar));
	NetworkSendPacket(kMessageTypeCarID,&message,sizeof(tCarIDMessage),kMessagePriorityHigh,kMessageSendToAllButSelf);
}

//adds the line in str to the chat buffer, and scrolls buffer if necessary
void ChatBufferInsert(tChatMessage *msg,tChatBuffer *cb)
{
	msg->recvTime=TimeGetSeconds();
	if(msg->str[0]!='\0')
	{
		if(cb->chatBufferLines<kChatBufferMaxLines)
		{
			cb->chatBuffer[cb->chatBufferLines]=*msg;
			cb->chatBufferLines++;
		}
		else //scrolling is needed
		{
			for(int i=0;i<kChatBufferMaxLines-1;i++)
				cb->chatBuffer[i]=cb->chatBuffer[i+1];
			cb->chatBuffer[kChatBufferMaxLines-1]=*msg;			
		}
	}	
}

void GameInfoSend(tGameInfo *gInfo)
{
	strcpy(gInfo->environmentName,FileGetName(gInfo->environment));
	strcpy(gInfo->mapName,FileGetName(gInfo->map));
	for(int i=0;i<gInfo->numPlayers;i++)
		strcpy(gInfo->playerCarNames[i],FileGetName(gInfo->playerCars[i]));
}

void GameInfoReceive(tGameInfo *gInfo)
{
	gInfo->environment=FileGetReference(gInfo->environmentName);
	gInfo->map=FileGetReference(gInfo->mapName);
	for(int i=0;i<gInfo->numPlayers;i++)
		gInfo->playerCars[i]=FileGetReference(gInfo->playerCarNames[i]);
}


void MultiplayerMenuUpdate(tMultiplayerMenuData *userData,int host)
{
	tInterfaceMenuDescribtion *menu=userData->menu;
	if(userData->gInfo->inited)
		if(userData->gInfo->network&kGameInfoNetworkForcePlayerCar)
			userData->playerCar=userData->gInfo->playerCars[0];
	if(userData->playerCar!=kFileErr)
	{
		tCarDefinition *car=(tCarDefinition*)FileGetParsedDataPtr(userData->playerCar,kParserTypeCarDesc,sizeof(tCarDefinition));
		sprintf(menu->items[kMultiplayerSelectCar].label,"Car: \255#a\255%s...",car->carName);
	}
	else
		sprintf(menu->items[kMultiplayerSelectCar].label,"Car: \255#a\255Data File Mismatch");	
	if(userData->gInfo->inited)
	{
		if(userData->gInfo->map!=-1)
		{
			tMapInfo *mapInfo=(tMapInfo*)FileGetParsedDataPtr(userData->gInfo->map,kParserTypeMapInfoDesc,sizeof(tMapInfo));
			tEnvironment *env=(tEnvironment*)FileGetParsedDataPtr(userData->gInfo->environment,kParserTypeEnvironmentDesc,sizeof(tEnvironment));
			sprintf(menu->items[kMultiplayerSelectMap].label,"Map: \255#a\255%s (%s%s, %d Laps)...",mapInfo->name,userData->gInfo->reverse?"Reverse, ":"",env->name,userData->gInfo->numLaps);
		}
		else
			sprintf(menu->items[kMultiplayerSelectMap].label,"Map: \255#a\255Data File Mismatch");
		switch(userData->gInfo->arcade)
		{
			case 0: sprintf(menu->items[kMultiplayerGameMode].label,"Game Mode: \255#a\255Simulation"); break;
			case 1: sprintf(menu->items[kMultiplayerGameMode].label,"Game Mode: \255#a\255Arcade"); break;
			case 2: sprintf(menu->items[kMultiplayerGameMode].label,"Game Mode: \255#a\255Turbo Arcade"); break;
		}
	}
	else
		sprintf(menu->items[kMultiplayerSelectMap].label,"Map: ");		
	if(!host)
	{
		if(userData->gInfo->network&kGameInfoNetworkForcePlayerCar)
			menu->items[kMultiplayerSelectCar].flags|=kInterfaceMenuItemDisabled;
		else
			menu->items[kMultiplayerSelectCar].flags&=~kInterfaceMenuItemDisabled;
		
		
		if(gConfirmedVersion!=userData->gInfo->playerVersions[userData->playerID])
		{
			menu->items[kMultiplayerStartGame].flags|=kInterfaceMenuItemDisabled;
			float time=TimeGetSeconds();
			if(time-userData->lastSendTime>0.5)
			{
				int reply[2];
				reply[0]=gConfirmedVersion;
				reply[1]=userData->playerID;
				S32ChangeToNet(reply[0]);
				S32ChangeToNet(reply[1]);
				if(gConfirmedVersion!=-1)
					NetworkSendPacket(kMessageTypeVersionConfirmed,reply,sizeof(int)*2,kMessagePriorityHigh,kMessageSendToHostOnly);
				else
					NetworkSendPacket(kMessageTypeVersionDenied,reply,sizeof(int)*2,kMessagePriorityHigh,kMessageSendToHostOnly);
				userData->lastSendTime=time;
			}
		}
		else
			menu->items[kMultiplayerStartGame].flags&=~kInterfaceMenuItemDisabled;
		
		if(userData->gInfo->playerVersions[userData->playerID]!=userData->gInfo->version)
			strcpy(menu->items[kMultiplayerStartGame].label,"Ready to Race: \255#a\255No");		
		else
			strcpy(menu->items[kMultiplayerStartGame].label,"Ready to Race: \255#a\255Yes");		
	}	
	else
	{
		int startable=true;		
		int compatible=true;		
		for(int i=1;i<userData->gInfo->numNetPlayers;i++)
			if(userData->gInfo->playerVersions[i]!=userData->gInfo->version)
			{
				startable=false;	
//				if(userData->gInfo->playerVersions[i]==-1)
//					compatible=false;
			}
//		if(!compatible)
//			strcpy(menu->items[kMultiplayerStartGame].label,"Data Files don't match");
//		else
	 	if(!startable)
			strcpy(menu->items[kMultiplayerStartGame].label,"Send Ready Notification");
		else
			strcpy(menu->items[kMultiplayerStartGame].label,"Start Game");
	}	
}


void FillAIPlayers(tMultiplayerMenuData *userData,tGameInfo *gInfo)
{
	int numCars=gInfo->numNetPlayers+gConfig->numEnemies;
	if(numCars>kMaxPlayers)numCars=kMaxPlayers;
	
	for(int i=gInfo->numNetPlayers;i<numCars;i++)
	{
		if(userData->gInfo->network&kGameInfoNetworkForcePlayerCar)
		{
			tCarDefinition *car=(tCarDefinition*)FileGetParsedDataPtr(userData->playerCar,kParserTypeCarDesc,sizeof(tCarDefinition));
			userData->gInfo->playerCars[i]=userData->playerCar;
			userData->gInfo->playerColors[i]=(userData->playerColor+i)%car->numColors;
		}
		else
		{
			gInfo->playerCars[i]=gConfig->opponentCars[i-gInfo->numNetPlayers];
			gInfo->playerColors[i]=gConfig->opponentColors[i-gInfo->numNetPlayers];
		}
		gInfo->netID[i]=0;
		gInfo->ping[i]=0;
		gInfo->playerVersions[i]=0;
		strcpy(gInfo->playerNames[i],"AI Driver");
	}
	gInfo->numPlayers=numCars;
}


int NetworkHostCallback(tMultiplayerMenuData *userData)
{
	int showLonelyMessage;
	NetworkAdvertiseIdle(userData->gInfo,&showLonelyMessage,userData->locked,false);
	if(showLonelyMessage)
	{
		tChatMessage m;
		m.flags=kChatFlagSystem;
		sprintf(m.str,"### Players from the Internet may be unable to join you because of the way your router/network is configured.");
		PlayInterfaceSound(FileGetReference("chat.wav"));
		ChatBufferInsert(&m,&userData->cb);
	}

	GameInfoSend(userData->gInfo);
	GatherPingStats(userData->gInfo);
	float time=TimeGetSeconds();
	if(time-userData->lastSendTime>1.0)
	{
		userData->lastSendTime=time;
		ConvertGInfoSwap(userData->gInfo);
		NetworkSendPacket(kMessageTypeGameInfo,userData->gInfo,sizeof(tGameInfo),kMessagePriorityHigh,kMessageSendToAllButSelf);
		ConvertGInfoSwap(userData->gInfo);
	}
	
	tChatMessage chatReceive;
	WaitForPlayers(userData->gInfo,&chatReceive);
	userData->gInfo->playerCars[0]=userData->playerCar;
	tCarDefinition *car=(tCarDefinition*)FileGetParsedDataPtr(userData->playerCar,kParserTypeCarDesc,sizeof(tCarDefinition));
	if(userData->gInfo->network&kGameInfoNetworkForcePlayerCar)
		for(int i=1;i<userData->gInfo->numNetPlayers;i++)
		{
			userData->gInfo->playerCars[i]=userData->playerCar;
			userData->gInfo->playerColors[i]=(userData->playerColor+i)%car->numColors;
		}
	strcpy(userData->gInfo->playerNames[0],gConfig->playerName);
	userData->gInfo->playerColors[0]=userData->playerColor;
	ChatBufferInsert(&chatReceive,&userData->cb);

	FillAIPlayers(userData,userData->gInfo);
	
	MultiplayerMenuUpdate(userData,true);
	return -1;
}

void CheckGInfoVersion(tMultiplayerMenuData *userData,int playerID)
{
	int ok=true;
	for(int i=0;i<userData->gInfo->numPlayers;i++)
		if(userData->gInfo->playerCars[i]==kFileErr)
		{
			tChatMessage m;
			m.flags=kChatFlagSystem;
			if(playerID!=i)
				sprintf(m.str,"### You don't have the data file for the car used by player %d.",i);
			else
				sprintf(m.str,"### The other players don't have the data file used by your car.");			
			PlayInterfaceSound(FileGetReference("chat.wav"));
			ChatBufferInsert(&m,&userData->cb);
			ok=false;
		}
	if(userData->gInfo->environment==kFileErr)
	{
		tChatMessage m;
		m.flags=kChatFlagSystem;
		sprintf(m.str,"### You don't have the data file for the weather used.");
		PlayInterfaceSound(FileGetReference("chat.wav"));
		ChatBufferInsert(&m,&userData->cb);
		ok=false;
	}
	if(userData->gInfo->map==kFileErr)
	{
		tChatMessage m;
		m.flags=kChatFlagSystem;
		sprintf(m.str,"### You don't have the data file for the track used.");
		PlayInterfaceSound(FileGetReference("chat.wav"));
		ChatBufferInsert(&m,&userData->cb);
		ok=false;
	}
	
//	int reply[2];
//	reply[0]=gInfo->version;
//	reply[1]=playerID;
	if(ok)
	{
//		NetworkSendPacket(kMessageTypeVersionConfirmed,reply,sizeof(int)*2,kMessagePriorityHigh,kMessageSendToHostOnly);
		gConfirmedVersion=userData->gInfo->version;
	}
	else
	{
//		NetworkSendPacket(kMessageTypeVersionDenied,reply,sizeof(int)*2,kMessagePriorityHigh,kMessageSendToHostOnly);
		gConfirmedVersion=-1;
	}
//	lastSendTime=TimeGetSeconds();
}

int NetworkClientCallback(tMultiplayerMenuData *userData)
{
	int exit=false;
	tChatMessage chatReceive;
	if(WaitForStartSignal(userData->gInfo,&chatReceive,&userData->playerID,&exit))
		return exit;
/*		if(exit)
			return kMultiplayerExit;
		else
			return kMultiplayerStartGameSignal;*/

	ChatBufferInsert(&chatReceive,&userData->cb);

	if(userData->gInfo->inited)
		GameInfoReceive(userData->gInfo);

	MultiplayerMenuUpdate(userData,false);
	return -1;
}

#define kPingGreen 0.1
#define kPingYellow 0.2
#define kPingRed 0.35

void RenderChatBuffer(tChatBuffer *cb,float xleft,float xright,float y,float textsize)
{
	int chatBufferDrawLines[kChatBufferMaxLines];
	int chatBufferTotalDrawLines=0;

	for(int i=0;i<cb->chatBufferLines;i++)
	{
		char str[256];
		int offset=0;
		chatBufferDrawLines[i]=0;
		do{
			strcpy(str,cb->chatBuffer[i].str+offset);
			int word=true;
			while(TextWidth(str,textsize)>(xright-xleft)*kTextXPos)
			{
				word=str[strlen(str)-1]==' ';
				str[strlen(str)-1]='\0';
			}
			if(!word)
			{
				int i=strlen(str)-1;
				while(str[i]!=' '&&i>0)
					i--;
				if(i>0)
					str[i]='\0';
			}
			offset+=strlen(str);
			chatBufferDrawLines[i]++;
			chatBufferTotalDrawLines++;
		}while(offset<strlen(cb->chatBuffer[i].str));		
	}
	
	int drawline=0;
	if(chatBufferTotalDrawLines>kChatBufferMaxLines)drawline=kChatBufferMaxLines-chatBufferTotalDrawLines;
	float t=TimeGetSeconds();

	for(int i=0;i<cb->chatBufferLines;i++)
	{
		char str[256];
		int offset=0;
		chatBufferDrawLines[i]=0;
		do{
			if(offset)
				sprintf(str,"\255#a\255%s",cb->chatBuffer[i].str+offset);
			else
				strcpy(str,cb->chatBuffer[i].str);
			int word=true;
			while(TextWidth(str,textsize)>(xright-xleft)*kTextXPos)
			{
				word=str[strlen(str)-1]==' ';
				str[strlen(str)-1]='\0';
			}
			if(!word)
			{
				int i=strlen(str)-1;
				while(str[i]!=' '&&i>0)
					i--;
				if(i>0)
					str[i]='\0';
			}
			offset+=strlen(str);
			float r=cb->chatBuffer[i].flags&kChatFlagSystem?0:1;
			float g=cb->chatBuffer[i].flags&kChatFlagAlert?0:1;
			float b=cb->chatBuffer[i].flags&(kChatFlagSystem+kChatFlagAlert)?0:1;
			if(drawline>=0)
				TextPrintfToBufferFormatedColored(Vector(xleft,y-drawline*0.08),textsize,kTextAlignLeft,r,g,b,(t-cb->chatBuffer[i].recvTime)*4,str);	
			drawline++;		
		}while(offset<strlen(cb->chatBuffer[i].str));		
	}
}

void NetworkRenderCallback(tMultiplayerMenuData *userData,int selection,void *menu)
{
	InterfaceRenderReplay(NULL,selection,menu);
	
	//print list of players
	if(userData->gInfo->inited)
	{
		for(int i=0;i<userData->gInfo->numPlayers;i++)
		{
			float a=userData->gInfo->playerVersions[i]==userData->gInfo->version?1:0.5;
			if(i>=userData->gInfo->numNetPlayers||i==0)a=1;
			if(userData->gInfo->playerCars[i]!=-1)
			{
				tCarDefinition *car=(tCarDefinition*)FileGetParsedDataPtr(userData->gInfo->playerCars[i],kParserTypeCarDesc,sizeof(tCarDefinition));
				if(i==userData->playerID)
					TextPrintfToBufferFormatedColored(Vector(0.4,0.6-i*0.15),0.028,kTextAlignLeft|kTextAutoCut,0,1,0,a,"%s: \255#n\255%s",userData->gInfo->playerNames[i],car->carName);
				else
					TextPrintfToBufferFormatedColored(Vector(0.4,0.6-i*0.15),0.028,kTextAlignLeft|kTextAutoCut,1,1,1,a,"%s: \255#n\255%s",userData->gInfo->playerNames[i],car->carName);
			}
			else if(!userData->gInfo->playerInited[i])
				TextPrintfToBufferFormatedColored(Vector(0.4,0.6-i*0.15),0.028,kTextAlignLeft|kTextAutoCut,0.8,0,0,0.5,"%s: \255#n\255Joining",userData->gInfo->playerNames[i]);			
			else
				TextPrintfToBufferFormatedColored(Vector(0.4,0.6-i*0.15),0.028,kTextAlignLeft|kTextAutoCut,0.8,0,0,0.5,"%s: \255#n\255Data File Mismatch",userData->gInfo->playerNames[i]);			
			float r,g,b;
			if(userData->gInfo->ping[i]<kPingGreen) {r=0;g=1;b=0;}
			else if(userData->gInfo->ping[i]<kPingYellow) {r=(userData->gInfo->ping[i]-kPingGreen)/(kPingYellow-kPingGreen);g=1;b=0;}
			else if(userData->gInfo->ping[i]<kPingRed) {r=1;g=1-(userData->gInfo->ping[i]-kPingYellow)/(kPingRed-kPingYellow);b=0;}
			else {r=1;g=0;b=0;}
			TextPrintfToBufferFormatedColored(Vector(0.45,0.54-i*0.15),0.022,kTextAlignLeft,r,g,b,1,"Ping: %3.0fms",userData->gInfo->ping[i]*1000.0);
			//TextPrintfToBufferFormatedColored(Vector(0.55,0.54-i*0.15),0.022,kTextAlignLeft,r,g,b,1,"Ping: %3.0fms",userData->gInfo->ping[i]*1000.0);
		}
	}
	else
		TextPrintfToBufferFormated(Vector(-0.85,0.6),0.03,kTextAlignLeft,"Waiting for Data...");	

	RenderChatBuffer(&userData->cb,-0.85,0.3,0.6,0.03);
}

enum{
	kHostControlAICars,
	kHostControlForcePlayerCar,
	kHostControlKickPlayer,
	kHostControlTrackerName,
	kHostControlPassword,
	kHostControlLockGame,
	kHostControlReturn,
	kHostControlNumOptions
};


void MultiplayerKickPlayer(tMultiplayerMenuData *userData)
{
	if(userData->gInfo->numNetPlayers<2)
		return;
		
	tInterfaceMenuDescribtion menu;
	InterfaceInitMenu(&menu,userData->gInfo->numNetPlayers,"Select a Player to Kick:");

	menu.TimerCallback=(int(*)(void*))NetworkHostCallback;
	menu.userData=userData;
	
	for(int i=1;i<userData->gInfo->numNetPlayers;i++)
		strcpy(menu.items[i-1].label,userData->gInfo->playerNames[i]);
	
	strcpy(menu.items[userData->gInfo->numNetPlayers-1].label,"Cancel");
	menu.items[userData->gInfo->numNetPlayers-2].lineSpacing*=1.5;
	
	int sel=InterfaceGetUserMenuSelection(&menu);
	InterfaceDisposeMenu(&menu);

	if(sel!=kInterfaceMenuEsc&&sel!=userData->gInfo->numNetPlayers-1)
		NetworkDisconnectPlayer(userData->gInfo->netID[sel+1]);
		
	tChatMessage m;
	m.flags=kChatFlagSystem;
	sprintf(m.str,"### Player %s was kicked by host.",userData->gInfo->playerNames[sel+1]);
	NetworkSendPacket(kMessageTypeChat,&m,sizeof(tChatMessage),kMessagePriorityHigh,kMessageSendToAll);					
}

void HostControl(tMultiplayerMenuData *userData)
{
	tInterfaceMenuDescribtion menu;
	InterfaceInitMenu(&menu,kHostControlNumOptions,"Host Controls");
	char itemStrings[][32]={"AI Cars...","Force Players to use host car: ","Kick Player...","Tracker Name: ","Password: ","Lock Game:","Return to chat"};
	for(int i=0;i<menu.numItems;i++)
		strcpy(menu.items[i].label,itemStrings[i]);

	char itemDesc[][256]={
		"Allows you to add or remove computer-controlled opponents to the game.",
		"This forces all players to use the same car as yours.",
		"You can use this to remove unwanted players from the game.",
		"Name for this game that other players will see.",
		"Password required to join the game. Leave empty to allow everyone to join.",
		"Allows you to lock the game, so no new players can join.",
		"Return to chat screen.",
		};
	
	for(int i=0;i<menu.numItems;i++)
		strcpy(menu.items[i].describtion,itemDesc[i]);
		
	menu.items[kHostControlReturn-1].lineSpacing*=1.5;
	menu.items[kHostControlLockGame].flags|=kInterfaceMenuItemArrowInput;
	menu.items[kHostControlForcePlayerCar].flags|=kInterfaceMenuItemArrowInput;
	menu.items[kHostControlTrackerName].flags|=kInterfaceMenuItemTypeable;
	menu.items[kHostControlPassword].flags|=kInterfaceMenuItemTypeable|kInterfaceMenuItemTypeHidden;
	if(userData->gInfo->numNetPlayers<2)
		menu.items[kHostControlKickPlayer].flags|=kInterfaceMenuItemDisabled;
	
	strcpy(menu.items[kHostControlTrackerName].type,gConfig->gameName);
	strcpy(menu.items[kHostControlPassword].type,gConfig->password);
	menu.TimerCallback=(int(*)(void*))(NetworkHostCallback);
	menu.RenderCallback=InterfaceRenderReplay;
	menu.userData=userData;
	sprintf(menu.items[kHostControlLockGame].label,"Lock Game: \255#a\255%s",userData->locked?"Yes":"No");
	sprintf(menu.items[kHostControlForcePlayerCar].label,"Force Players to use host car: \255#a\255%s",userData->gInfo->network&kGameInfoNetworkForcePlayerCar?"Yes":"No");
	
	int exit=false;
	InterfaceMenuZoomAnimation(&menu,0,true);
	do{
		sprintf(menu.items[kHostControlLockGame].label,"Lock Game: \255#a\255%s",userData->locked?"Yes":"No");
		sprintf(menu.items[kHostControlForcePlayerCar].label,"Force Players to use host car: \255#a\255%s",userData->gInfo->network&kGameInfoNetworkForcePlayerCar?"Yes":"No");
		int sel=InterfaceGetUserMenuSelection(&menu);
		switch(menu.initialSelection=(sel&kInterfaceMenuItemMask))
		{
			case kHostControlAICars:
				tFileRef opponents[5];
				UInt8 colors[5];
				int numOpponents;
				InterfaceSelectOpponentCars(&numOpponents,opponents,colors,(int(*)(void*))(NetworkHostCallback),userData);
				break;
			case kHostControlKickPlayer:
				MultiplayerKickPlayer(userData);
				break;
			case kHostControlLockGame:
				userData->locked=!userData->locked;
				break;				
			case kHostControlForcePlayerCar:
				userData->gInfo->network^=kGameInfoNetworkForcePlayerCar;
				userData->gInfo->version++;
				userData->gInfo->playerVersions[0]=userData->gInfo->version;
				break;				
			case kHostControlReturn:
			case kInterfaceMenuEsc:
				exit=true;
				break;
		}
	}while(!exit);
	strcpy(gConfig->gameName,menu.items[kHostControlTrackerName].type);
	strcpy(gConfig->password,menu.items[kHostControlPassword].type);
	NetworkChangePassword(gConfig->password);
	NetworkAdvertiseIdle(userData->gInfo,NULL,userData->locked,true);
	if(userData->locked)
		NetworkLockOutNewPlayers();
	else
		NetworkUnlockOutNewPlayers();
	InterfaceMenuZoomAnimation(&menu,menu.initialSelection,false);
	InterfaceDisposeMenu(&menu);	
}

void InitMultiPlayerMenu(tInterfaceMenuDescribtion *menu,int host,tMultiplayerMenuData *userData)
{
	InterfaceInitMenu(menu,kNumMultiplayerOptions,"Multiplayer");

	char itemStrings[][32]={"Chat Prompt: ","Car: ","Map: ","Game Mode: ","Host Controls...","Start Game","Exit Multiplayer"};
	for(int i=0;i<menu->numItems;i++)
	{
		strcpy(menu->items[i].label,itemStrings[i]);
		menu->items[i].size*=i<kMultiplayerStartGame?0.75:0.85;
		menu->items[i].lineSpacing*=i<kMultiplayerStartGame?0.75:0.85;
	}
	menu->items[kMultiplayerStartGame-1].lineSpacing*=1.5;
	if(!host)
	{
		menu->items[kMultiplayerSelectMap].flags|=kInterfaceMenuItemDisabled;
		menu->items[kMultiplayerHostControls].flags|=kInterfaceMenuItemDisabled;
		menu->items[kMultiplayerGameMode].flags|=kInterfaceMenuItemDisabled;
//		menu->items[kMultiplayerMiniGame].flags|=kInterfaceMenuItemDisabled;
		menu->items[kMultiplayerStartGame].flags|=kInterfaceMenuItemArrowInput;
	}
	menu->items[kMultiplayerSelectCar].flags|=kInterfaceMenuItemArrowInput;
	menu->items[kMultiplayerSelectMap].flags|=kInterfaceMenuItemArrowInput;
	menu->items[kMultiplayerGameMode].flags|=kInterfaceMenuItemArrowInput;
	menu->items[kMultiplayerChat].flags|=kInterfaceMenuItemTypeable;
	menu->items[kMultiplayerChat].maxTypeXPos=0.35;
	
	menu->background=FileGetReference("background-multiplayer.tif");
	menu->joinDisable=true;
	menu->itemsYPos=-0.1;
	menu->itemsXPos=-0.87;
	menu->TimerCallback=(int(*)(void*))(host?NetworkHostCallback:NetworkClientCallback);
	menu->RenderCallback=(void(*)(void*,int,void*))NetworkRenderCallback;
	menu->userData=userData;
}

void RunMiniGame(tGameInfo *gInfo)
{
	tGameInfo miniGInfo;
	InitGInfo(&miniGInfo);
	miniGInfo.map=FileGetReference("pit.mapinfo");
	miniGInfo.environment=FileGetReference("sunset.env");
	miniGInfo.demolition=true;
	miniGInfo.numPlayers=gInfo->numNetPlayers;
	miniGInfo.numNetPlayers=gInfo->numNetPlayers;
	miniGInfo.playerID=0;
	for(int i=0;i<gInfo->numNetPlayers;i++)
	{
		miniGInfo.playerCars[i]=FileGetReference("oldford.car");
		strcpy(miniGInfo.playerNames[i],gInfo->playerNames[i]);
	}
	miniGInfo.network=true;

	GameInfoSend(&miniGInfo);
	ConvertGInfoSwap(&miniGInfo);
	NetworkSendPacket(kMessageTypeStart,&miniGInfo,sizeof(tGameInfo),kMessagePriorityHigh,kMessageSendToAllButSelf);
	ConvertGInfoSwap(&miniGInfo);

	RunGame(&miniGInfo);	

	for(int i=1;i<miniGInfo.numPlayers;i++)
		if(miniGInfo.netID[i]==-1)
			gInfo->netID[i]=-1;

	FixGInfoForLeftPlayers(gInfo);

	ConvertGInfoSwap(gInfo);
	NetworkSendPacket(kMessageTypeGameInfo,gInfo,sizeof(tGameInfo),kMessagePriorityHigh,kMessageSendToAllButSelf);
	ConvertGInfoSwap(gInfo);
}

int SelectRaceMode(tFileRef map,int (*TimerCallback)(void*),void *userData);

int HandleMultiplayerMenuItem(tInterfaceMenuDescribtion *menu,tMultiplayerMenuData *userData,int item,int host)
{
	int callbackResponse=-1;
	switch(item&kInterfaceMenuItemMask)
	{
		case kMultiplayerGameMode:
			if(item&kInterfaceMenuLeftArrow){
				userData->gInfo->arcade--;
				if(userData->gInfo->arcade<0)
					userData->gInfo->arcade=2;//(item&kInterfaceMenuEasterEgg?2:1);
			}
			else if(item&kInterfaceMenuRightArrow)
			{
				userData->gInfo->arcade=(userData->gInfo->arcade+1);
				if(userData->gInfo->arcade>2)//(item&kInterfaceMenuEasterEgg?2:1))
					userData->gInfo->arcade=0;
			}
			else
			{
				int arcade=SelectRaceMode(userData->gInfo->map,(int(*)(void*))menu->TimerCallback,userData);
				if(arcade==-1)
					break;
				userData->gInfo->arcade=arcade;
			}
			userData->gInfo->version++;
			userData->gInfo->playerVersions[0]=userData->gInfo->version;
			break;

		case kMultiplayerSelectMap:
			if(item&kInterfaceMenuLeftArrow)
				SelectPrevMap(&userData->gInfo->map);
			else if(item&kInterfaceMenuRightArrow)
				SelectNextMap(&userData->gInfo->map);
			else
				InterfaceMapSelection(userData->gInfo,(int(*)(void*))menu->TimerCallback,userData);
			userData->gInfo->version++;
			userData->gInfo->playerVersions[0]=userData->gInfo->version;
			break;
			
		case kMultiplayerSelectCar:
			if(userData->gInfo->network&kGameInfoNetworkForcePlayerCar&&userData->playerID>0)
				break;
			if(item&kInterfaceMenuLeftArrow)
				SelectPrevCar(&userData->playerCar,&userData->playerColor,true);
			else if(item&kInterfaceMenuRightArrow)
				SelectNextCar(&userData->playerCar,&userData->playerColor,true);
			else{
				InterfaceCarSelection(&userData->playerCar,kCarSelectionQuickMode,&userData->playerColor,(int(*)(void*))menu->TimerCallback,userData,&callbackResponse);
				if(callbackResponse!=-1)
					return HandleMultiplayerMenuItem(menu,userData,callbackResponse,host);
			}
			if(userData->playerID==0)
			{
				userData->gInfo->version++;
				userData->gInfo->playerVersions[0]=userData->gInfo->version;
			}
			else if(userData->gInfo->playerCars[userData->playerID]!=userData->playerCar||userData->gInfo->playerColors[userData->playerID]!=userData->playerColor)
				UpdatePlayerCar(userData);
				
			gConfig->lastCar=userData->playerCar;
			gConfig->lastColor=userData->playerColor;
			break;
			
		case kMultiplayerHostControls:
			HostControl(userData);
			break;
			
		case kMultiplayerChat:
			if(menu->items[kMultiplayerChat].type[0])
			{
				tChatMessage m;
				m.flags=0;
				sprintf(m.str,"%s:\255#a\255 %s",gConfig->playerName,menu->items[kMultiplayerChat].type);
				NetworkSendPacket(kMessageTypeChat,&m,sizeof(tChatMessage),kMessagePriorityHigh,kMessageSendToAll);					
				menu->items[kMultiplayerChat].type[0]='\0';
			}
			break;
			
		case kMultiplayerStartGame:
			if(host)
			{
				int startable=true;
				for(int i=0;i<userData->gInfo->numNetPlayers;i++)
					if(userData->gInfo->playerVersions[i]!=userData->gInfo->version)
					{
						tChatMessage m;
						m.flags=kChatFlagAlert;
						strcpy(m.str,"### Please Select Ready, Host wants to start game!!!");
						NetworkSendPacket(kMessageTypeChat,&m,sizeof(m),kMessagePriorityHigh,userData->gInfo->netID[i]);					
						startable=false;
					}
				return startable;
			}
			else if(userData->gInfo->inited)
			{
				if(userData->gInfo->playerVersions[userData->playerID]!=userData->gInfo->version)
					CheckGInfoVersion(userData,userData->playerID);
				else
				{
					int reply[2];
					reply[0]=userData->gInfo->version;
					reply[1]=userData->playerID;
					S32ChangeToNet(reply[0]);
					S32ChangeToNet(reply[1]);
					NetworkSendPacket(kMessageTypeVersionDenied,reply,sizeof(int)*2,kMessagePriorityHigh,kMessageSendToHostOnly);
					gConfirmedVersion=-1;
				}			
			}
			break;

		/*case kMultiplayerMiniGame:
			if(host)
				RunMiniGame(userData->gInfo);
			break;*/
			
		case kMultiplayerExit:
		case kInterfaceMenuEsc:
			return -1;
		case kMultiplayerFailed:
			return -2;
		case kMultiplayerKicked:
			return -3;
		case kMultiplayerStartGameSignal:
			return 1;
	}
	return 0;
}

void MultiplayerMenu(int host,tMultiplayerMenuData *userData,char* exitMessage)
{
	tInterfaceMenuDescribtion menu;
	userData->menu=&menu;
	InitMultiPlayerMenu(&menu,host,userData);
	
	int exit=false;
	while(!exit)
	{
		int start=false;
		int zoom=true;
		do{
			if(zoom)
				InterfaceMenuZoomAnimation(&menu,-1,true);
			
			int item=InterfaceGetUserMenuSelection(&menu);
			int response=HandleMultiplayerMenuItem(&menu,userData,item,host);
			menu.initialSelection=(item&kInterfaceMenuItemMask);
			if(menu.initialSelection!=kMultiplayerChat||menu.initialSelection!=kMultiplayerStartGame||item&(kInterfaceMenuLeftArrow|kInterfaceMenuRightArrow))
				zoom=false;
			else 
				zoom=true;
			
			if(response<0)
			{
				exit=true;
				if(exitMessage)
				{
					strcpy(exitMessage,"");
					if(response==-2)
						strcpy(exitMessage,"Connection lost.");
					if(response==-3)
						strcpy(exitMessage,"You were kicked out of the game.");
				}
			}
			else if(response>0)start=true;
		}while(!exit&&!start);
		if(start)
		{
			if(host)
				StartNetGame(userData->gInfo);
			else{
				GameInfoReceive(userData->gInfo);
				userData->gInfo->playerID=userData->playerID;
			}
			if(userData->gInfo->numNetPlayers>1)
			{
				NetworkLockOutNewPlayers();
				if(host)
					NetworkAdvertiseIdle(userData->gInfo,NULL,true,true);
			}
			RunGame(userData->gInfo);
			NetworkUnlockOutNewPlayers();
			if(host)
				FixGInfoForLeftPlayers(userData->gInfo);
		}
	}
	InterfaceDisposeMenu(&menu);
}

//hosts a network game
void InterfaceHostNetworkGame()
{
	NetworkClearPacketQueue();
	
	LocalTrackerStartAdvertising(gConfig->gameName);

	tGameInfo gInfo;
	HostNetGame(gConfig->maxPlayers,gConfig->password);	

	InitGInfo(&gInfo);

	if(gConfig->lastCar==-1||gConfig->lastCar==0)
	{
		UInt8 color;
		SelectNextCar(&gConfig->lastCar,&color,true);
		gConfig->lastColor=color;
	}
	gInfo.network=true;
	gInfo.playerCars[0]=gConfig->lastCar;
	gInfo.playerColors[0]=gConfig->lastColor;
	gInfo.map=gConfig->lastRoad;
	gInfo.numLaps=gConfig->lastLaps;
	gInfo.environment=gConfig->lastEnv;
	gInfo.inited=true;
	gInfo.arcade=gConfig->arcade;
	
	tMultiplayerMenuData userData;
	userData.gInfo=&gInfo;
	userData.lastSendTime=0;
	userData.playerCar=gConfig->lastCar;
	userData.playerID=0;
	userData.locked=false;
	userData.t=INFINITY;
	userData.playerColor=gConfig->lastColor;
	userData.cb.chatBufferLines=0;
	
	MultiplayerMenu(true,&userData,NULL);

	NetworkStopAdvertising();
	LocalTrackerStopAdvertising();
	ExitNetGame();
}


enum{
	kHostOptionsTracker,
	kHostOptionsTrackerName,
	kHostOptionsMaxPlayers,
	kHostOptionsPassword,
	kHostOptionsPlayerName,
	kHostOptionsStart,
	kHostOptionsCancel,
	kNumHostOptions
};

void InterfaceSetupHost()
{
	tInterfaceMenuDescribtion menu;
	InterfaceInitMenu(&menu,kNumHostOptions,"Host Game");
	
	menu.RenderCallback=InterfaceRenderReplay;
	strcpy(menu.items[kHostOptionsCancel].label,"Return to previous menu");
	strcpy(menu.items[kHostOptionsStart].label,"Start Hosting");
	strcpy(menu.items[kHostOptionsTrackerName].label,"Game Name: ");
	strcpy(menu.items[kHostOptionsPassword].label,"Game Password: ");
	strcpy(menu.items[kHostOptionsPlayerName].label,"Local Player Name: ");

	char itemDesc[][256]={
		"Advertises the game to other players on the internet tracker.",
		"Name for this game that other players will see.",
		"Maximum number of players who can join the game.",
		"Password required to join the game. Leave empty to allow everyone to join.",
		"Your player name.",
		"Start hosting the game.",
		"Return to multiplayer selection menu.",
		};
	for(int i=0;i<menu.numItems;i++)
		strcpy(menu.items[i].describtion,itemDesc[i]);

	for(int i=0;i<kHostOptionsStart;i++)
	{
		menu.items[i].size*=0.75;
		menu.items[i].lineSpacing*=0.75;
	}
	menu.items[kHostOptionsStart-1].lineSpacing*=1.5;
	
	menu.items[kHostOptionsTracker].flags|=kInterfaceMenuItemArrowInput;
	menu.items[kHostOptionsTrackerName].flags|=kInterfaceMenuItemTypeable;
	menu.items[kHostOptionsMaxPlayers].flags|=kInterfaceMenuItemArrowInput;
	strcpy(menu.items[kHostOptionsTrackerName].type,gConfig->gameName);
	menu.items[kHostOptionsPassword].flags|=kInterfaceMenuItemTypeable;
	menu.items[kHostOptionsPassword].flags|=kInterfaceMenuItemTypeable|kInterfaceMenuItemTypeHidden;
	strcpy(menu.items[kHostOptionsPassword].type,gConfig->password);
	menu.items[kHostOptionsPlayerName].flags|=kInterfaceMenuItemTypeable;
	menu.items[kHostOptionsPlayerName].maxTypeLength=kMaxNameLength-1;
	strcpy(menu.items[kHostOptionsPlayerName].type,gConfig->playerName);

	int exit=false;
	int zoom=false;
	do{
		sprintf(menu.items[kHostOptionsTracker].label,"List Game on Tracker: \255#a\255%s",gConfig->trackerEnable?"On":"Off");
		sprintf(menu.items[kHostOptionsMaxPlayers].label,"Max Number of Players: \255#a\255%d",gConfig->maxPlayers);
		if(!zoom){
			InterfaceMenuZoomAnimation(&menu,-1,true);
			zoom=true;
		}
		
		int sel=InterfaceGetUserMenuSelection(&menu);
		switch(menu.initialSelection=(sel&kInterfaceMenuItemMask))
		{
			case kHostOptionsTracker:
				gConfig->trackerEnable=!gConfig->trackerEnable;
				break;
			
			case kHostOptionsMaxPlayers:
				if(sel&kInterfaceMenuLeftArrow){
					if(gConfig->maxPlayers>2)
						gConfig->maxPlayers--;
				}
				else
					if(gConfig->maxPlayers<kMaxPlayers)
						gConfig->maxPlayers++;
				break;
								
			case kHostOptionsPassword:
			case kHostOptionsTrackerName:
			case kHostOptionsPlayerName:
			case kHostOptionsStart:
			case kInterfaceMenuOK:
				strcpy(gConfig->gameName,menu.items[kHostOptionsTrackerName].type);
				strcpy(gConfig->password,menu.items[kHostOptionsPassword].type);
				strcpy(gConfig->playerName,menu.items[kHostOptionsPlayerName].type);
				InterfaceHostNetworkGame();
			case kHostOptionsCancel:
			case kInterfaceMenuEsc:
				exit=true;
				break;
		}
	}while(!exit);
	WriteOutFile(FileGetReference("config.cfg"),gConfig,kParserTypeConfigDesc);
	InterfaceDisposeMenu(&menu);
}

//join a network game
void InterfaceJoinNetworkGame(char *address,char *exitMessage)
{
	int err,id;
	char errorString[256];

	InterfaceDrawStrings("Connecting to Network Game...",address,kFileErr);
	if(gConfig->lastCar==-1||gConfig->lastCar==0)
	{
		UInt8 color;
		SelectNextCar(&gConfig->lastCar,&color,true);
		gConfig->lastColor=color;
	}

	//try to join a game at the given address
	if(JoinNetGame(address,&id,gConfig->lastCar,errorString))
	{
		tGameInfo gInfo;
		tMultiplayerMenuData userData;
		gInfo.inited=false;
		userData.gInfo=&gInfo;
		userData.lastSendTime=0;
		userData.playerColor=gConfig->lastColor;
		userData.playerCar=gConfig->lastCar;
		userData.playerID=id;
		userData.cb.chatBufferLines=0;
		gConfirmedVersion=0;
		
		MultiplayerMenu(false,&userData,exitMessage);
	}
	else
		InterfaceDisplayMessage(-1,"Unable To Connect.",errorString);
}

int gLastLocalGameListUpdate=0;

#define kMaxGames 256
#define kMaxPlayers 256

typedef struct{
	tInterfaceMenuDescribtion *menu;
	tGameListEntry games[kMaxGames];
	int numGames;
	tGameListEntry localGames[kMaxGames];
	int numLocalGames;
	tGameListPlayerEntry players[kMaxPlayers];
	int numPlayers;	
	int inited;
	int numGamesInMenu;
	tChatBuffer cb;
} tFindGamesMenuData;

enum{
	kLobbyHostGame,
	kLobbyJoinIP,
	kLobbyChat,
	kNumLobbyMenuItems
};

int FindGamesCallback(tFindGamesMenuData *userData)
{
	int updated;
	NetworkSearchGames(&updated,userData->games,&userData->numGames,kMaxGames,userData->players,&userData->numPlayers,kMaxPlayers);
	if(gLocalGameListUpdate!=gLastLocalGameListUpdate)
	{
		updated=true;	
		userData->numLocalGames=gNumLocalGames;
		if(userData->numLocalGames>kMaxGames)
			userData->numLocalGames=kMaxGames;
		MemoryMove(userData->localGames,gLocalGames,userData->numLocalGames*sizeof(tGameListEntry));
		gLastLocalGameListUpdate=gLocalGameListUpdate;
	}
	if(updated)
		userData->inited=true;
	//if(userData->numGames!=userData->numGamesInMenu)
	{
		if(userData->numGames==0&&userData->numLocalGames==0)
		{
			if(userData->inited)
				strcpy(userData->menu->items[kNumLobbyMenuItems].label,"<no games available>");
			else
				strcpy(userData->menu->items[kNumLobbyMenuItems].label,"<connecting to tracker>");
			strcpy(userData->menu->items[kNumLobbyMenuItems].describtion,"");
			userData->menu->numItems=kNumLobbyMenuItems+1;
			userData->menu->items[kNumLobbyMenuItems].flags|=kInterfaceMenuItemDisabled;
		}
		else
		{			
			for(int i=0;i<userData->numLocalGames;i++)
			{
				sprintf(userData->menu->items[kNumLobbyMenuItems+i].label,"%s \255rendezvous.pct\255",userData->localGames[i].name);
				userData->menu->items[kNumLobbyMenuItems+i].flags=0;
			}
			for(int i=userData->numLocalGames;i<userData->numGames+userData->numLocalGames;i++)
			{
				if(userData->games[i-userData->numLocalGames].loaded)
				{
					userData->menu->items[kNumLobbyMenuItems+i].flags=0;
					if(userData->games[i-userData->numLocalGames].numPlayers>0)
					{
						sprintf(userData->menu->items[kNumLobbyMenuItems+i].label,"%s (%d/%d) %s",userData->games[i-userData->numLocalGames].name
								,userData->games[i-userData->numLocalGames].numNetPlayers?userData->games[i-userData->numLocalGames].numNetPlayers:userData->games[i-userData->numLocalGames].numPlayers
								,userData->games[i-userData->numLocalGames].maxPlayers
								,userData->games[i-userData->numLocalGames].password?" \255lock.pct\255":"");
						if(userData->games[i-userData->numLocalGames].numNetPlayers>=userData->games[i-userData->numLocalGames].maxPlayers)
							userData->menu->items[kNumLobbyMenuItems+i].flags|=kInterfaceMenuItemDisabled;
					}
					else
						strcpy(userData->menu->items[kNumLobbyMenuItems+i].label,userData->games[i-userData->numLocalGames].host);				
					if(userData->games[i-userData->numLocalGames].locked)
						userData->menu->items[kNumLobbyMenuItems+i].flags|=kInterfaceMenuItemDisabled;
				}
				else
				{
					strcpy(userData->menu->items[kNumLobbyMenuItems+i].label,"<loading>");								
					userData->menu->items[kNumLobbyMenuItems+i].flags|=kInterfaceMenuItemDisabled;
				}
			}
			userData->menu->numItems=userData->numGames+userData->numLocalGames+kNumLobbyMenuItems;
		}
		userData->numGamesInMenu=userData->numGames+userData->numLocalGames;
	}
	
	tChatMessage msg;
	if(TrackerMessageReceive(msg.str))
	{
		msg.flags=0;
		ChatBufferInsert(&msg,&userData->cb);
		PlayInterfaceSound(FileGetReference("chat.wav"));
	}
	
	return -1;
}

void FindGamesRenderCallback(tFindGamesMenuData *userData,int selection,void *menu)
{
	InterfaceRenderReplay(NULL,selection,menu);

	if(selection-kNumLobbyMenuItems>=0&&selection-kNumLobbyMenuItems<userData->numGames+userData->numLocalGames)
	{
		tGameListEntry *game=selection+kNumLobbyMenuItems<userData->numLocalGames?userData->localGames+selection-kNumLobbyMenuItems:userData->games+selection-kNumLobbyMenuItems-userData->numLocalGames;
		if(!game->loaded)	
			return;
			
		TextPrintfToBufferFormated(Vector(0.28,0.59),0.028,kTextAlignLeft,"redline://%s",game->host);
		
		if(game->numPlayers>0)
		{
			switch(game->arcade)
			{
				case 0:
					TextPrintfToBufferFormated(Vector(0.28,0.51),0.028,kTextAlignLeft,"Game Mode: \255#a\255Simulation");
					break;
				case 1:
					TextPrintfToBufferFormated(Vector(0.28,0.51),0.028,kTextAlignLeft,"Game Mode: \255#a\255Arcade");
					break;
				case 2:
					TextPrintfToBufferFormated(Vector(0.28,0.51),0.028,kTextAlignLeft,"Game Mode: \255#a\255Turbo Arcade");
					break;
			}
			TextPrintfToBufferFormated(Vector(0.28,0.43),0.028,kTextAlignLeft+kTextAutoCut,"Track: \255#a\255%s",game->map);
			TextPrintfToBufferFormated(Vector(0.28,0.35),0.028,kTextAlignLeft,"Password Protection: \255#a\255%s",game->password?"On":"Off");

			TextPrintfToBufferFormated(Vector(0.28,-0.07),0.028,kTextAlignLeft,"Players in Game: \255#a\255%d/%d",game->numPlayers,game->maxPlayers);
			for(int j=0;j<game->numPlayers;j++)
			{
				TextPrintfToBufferFormated(Vector(0.28,-0.18-j*0.11),0.028,kTextAlignLeft+kTextAutoCut,"%s: \255#a\255%s",game->players[j].name,game->players[j].car);
				TextPrintfToBufferFormated(Vector(0.97,-0.25-j*0.11),0.022,kTextAlignRight,game->players[j].location);
			}	
			if(game->locked)
				TextPrintfToBufferFormated(Vector(0.28,0.24),0.028,kTextAlignLeft+kTextAutoWrap,"Game is in progress or not accepting new players.");
		/*	else if(game->joinState==kJoinStateOK)
				TextPrintfToBufferFormated(Vector(0.28,0.24),0.028,kTextAlignLeft+kTextAutoWrap,"Joining this game shouldn't cause any troubles.");
			else if(game->joinState==kJoinStateReggie)
				TextPrintfToBufferFormated(Vector(0.28,0.24),0.028,kTextAlignLeft+kTextAutoWrap,"We'll need to pull off some tricks, but joining will probably work.");
		*/	else if(game->joinState==kJoinStateFail)
				TextPrintfToBufferFormatedColored(Vector(0.28,0.24),0.028,kTextAlignLeft+kTextAutoWrap,1,0,0,1,"Due to the way the networks are set up, you probably won't be able to join!");
		}
		else
		{
			TextPrintfToBufferFormated(Vector(0.28,0.51),0.028,kTextAlignLeft,"Game Mode: \255#a\255n/a");
			TextPrintfToBufferFormated(Vector(0.28,0.43),0.028,kTextAlignLeft,"Track: \255#a\255n/a");
			TextPrintfToBufferFormated(Vector(0.28,0.35),0.028,kTextAlignLeft,"Password Protection: \255#a\255n/a");
		}
	}
	else
	{
		TextPrintfToBufferFormated(Vector(0.28,0.65),0.028,kTextAlignLeft,"Players idling on tracker: \255#a\255%d",userData->numPlayers);
		for(int j=0;j<userData->numPlayers;j++)
		{
			TextPrintfToBufferFormated(Vector(0.28,0.52-j*0.11),0.028,kTextAlignLeft+kTextAutoCut,"%s",userData->players[j].name);
			TextPrintfToBufferFormated(Vector(0.97,0.45-j*0.11),0.022,kTextAlignRight,userData->players[j].location);
		}	
	}
//	TextPrintfToBufferFormated(Vector(-0.95,0.65),0.028,kTextAlignLeft,"Active Games: %d",userData->numGames);
//	TextPrintfToBufferFormated(Vector( 0.22,0.65),0.028,kTextAlignRight,"Players idling on tracker: %d",userData->numPlayers);
	
	RenderChatBuffer(&userData->cb,-0.9,0.27,-0.38,0.03);
}

void ManuallyJoinGame()
{
	NetworkPreSession();
	char address[256]="";
	char exitMessage[256]="";
	if(InterfaceGetUserInputString("Enter IP Address:",address,256,false,false))
		InterfaceJoinNetworkGame(address,exitMessage);	
	ExitNetGame();
	if(*exitMessage)
		InterfaceDisplayMessage(kFileErr,"Disconnected",exitMessage);
}

void InterfaceMultiplayerLobby()
{
	tFindGamesMenuData userData;
	
	tInterfaceMenuDescribtion menu;
	InterfaceInitMenu(&menu,kNumLobbyMenuItems+kMaxGames,"Multiplayer Lobby");
	strcpy(menu.items[kLobbyHostGame].label,"Host Game...");
	strcpy(menu.items[kLobbyJoinIP].label,"Enter IP Address...");
	strcpy(menu.items[kLobbyChat].label,"Chat Promt: ");
	menu.items[kLobbyChat].flags=kInterfaceMenuItemTypeable;
	menu.items[kLobbyChat].maxTypeXPos=0.2;
	for(int i=0;i<kNumLobbyMenuItems;i++){
		menu.items[i].flags|=kInterfaceMenuItemFixedPos;
		menu.items[i].fixedX=-0.9;
		menu.items[i].fixedY=-0.05-i*0.1;
	}
	for(int i=0;i<kNumLobbyMenuItems+kMaxGames;i++){
		menu.items[i].size=0.03;
		menu.items[i].lineSpacing=0.1;
	}

	NetworkPreSession();
	LocalTrackerSearchGamesInit();
	gLastLocalGameListUpdate=0;
	
	menu.userData=&userData;
	menu.scrollEnable=true;
	menu.itemsYPos+=0.4;
	userData.menu=&menu;
	userData.numGamesInMenu=0;
	userData.numGames=0;
	userData.numPlayers=0;
	userData.numLocalGames=0;
	userData.inited=false;
	userData.cb.chatBufferLines=0;
	
	menu.TimerCallback=(int(*)(void*))FindGamesCallback;
	menu.RenderCallback=(void(*)(void*,int,void*))FindGamesRenderCallback;
	menu.numItems=kNumLobbyMenuItems;
	menu.joinDisable=true;
	menu.background=FileGetReference("background-tracker.tif");
				
	do{
		switch(menu.initialSelection=InterfaceGetUserMenuSelection(&menu))
		{
			case kLobbyChat:
				if(*menu.items[kLobbyChat].type)
				{
					char msg[1024];
					sprintf(msg,"%s:\255#a\255 %s",gConfig->playerName,menu.items[kLobbyChat].type);
					TrackerMessageSend(msg);
					*menu.items[kLobbyChat].type='\0';
				}
				break;
			case kLobbyHostGame:
				LocalTrackerSearchGamesDone();
				NetworkSearchGamesDone();
				ExitNetGame();
				InterfaceSetupHost();
				LocalTrackerSearchGamesInit();
				NetworkPreSession();
				break;
			case kLobbyJoinIP:
				LocalTrackerSearchGamesDone();
				NetworkSearchGamesDone();
				ExitNetGame();
				ManuallyJoinGame();
				LocalTrackerSearchGamesInit();
				NetworkPreSession();
				break;
			default:
				if(menu.initialSelection!=kInterfaceMenuEsc)
				{
					NetworkSearchGamesStop();
					LocalTrackerSearchGamesDone();
					char exitMessage[256]="";
					char address[256]="";
					strcpy(address,
						menu.initialSelection-kNumLobbyMenuItems<userData.numLocalGames?
						userData.localGames[menu.initialSelection-kNumLobbyMenuItems].host
						:userData.games[menu.initialSelection-kNumLobbyMenuItems-userData.numLocalGames].host
					);
					InterfaceJoinNetworkGame(address,exitMessage);	
					if(*exitMessage)
						InterfaceDisplayMessage(kFileErr,"Disconnected",exitMessage);
					NetworkSearchGamesDone();
					ExitNetGame();
					LocalTrackerSearchGamesInit();
					NetworkPreSession();
				}
		}
	}while(menu.initialSelection!=kInterfaceMenuEsc);
	
	InterfaceDisposeMenu(&menu);

	LocalTrackerSearchGamesDone();
	NetworkSearchGamesDone();
	ExitNetGame();

//	NetworkSearchGamesStop();
/*	char exitMessage[256]="";
	if(sel!=kInterfaceMenuEsc)
	{
		char address[256]="";
		strcpy(address,sel<userData.numLocalGames?userData.localGames[sel].host:userData.games[sel-userData.numLocalGames].host);
		InterfaceJoinNetworkGame(address,exitMessage);	
	}
	NetworkSearchGamesDone();
	if(*exitMessage)
		InterfaceDisplayMessage(kFileErr,"Disconnected",exitMessage);*/
}

void DirectJoinGame(char *address)
{
	NetworkPreSession();
	char exitMessage[256]="";
	InterfaceJoinNetworkGame(address,exitMessage);	
	ExitNetGame();
	if(*exitMessage)
		InterfaceDisplayMessage(kFileErr,"Disconnected",exitMessage);

}
/*
enum{
	kInterfaceMultiplayerJoinGameTracker,
	kInterfaceMultiplayerJoinGameManually,
	kInterfaceMultiplayerHostGame,
	kInterfaceMultiplayerReturn,
	kNumInterfaceMultiplayerOptions
};

void InterfaceMultiplayerMenu()
{
	tInterfaceMenuDescribtion menu;
	InterfaceInitMenu(&menu,kNumInterfaceMultiplayerOptions,"Multiplayer");

	menu.items[kInterfaceMultiplayerHostGame].lineSpacing*=1.5;
	
	char itemDesc[][256]={
		"Automatically search for multiplayer games you can join on the internet and your local network.",
		"Manually join a game by entering the server's IP-address.",
		"Start your own game for others to join.",
		"Go back to main menu."
		};
	for(int i=0;i<menu.numItems;i++)
		strcpy(menu.items[i].describtion,itemDesc[i]);

	char itemStrings[][32]={"Search for games...","Enter IP-Address...","Host Game...","Return to previous menu"};
	for(int i=0;i<menu.numItems;i++)
		strcpy(menu.items[i].label,itemStrings[i]);

	for(int i=0;i<menu.numItems;i++)
	{
		menu.items[i].size*=1.15;
		menu.items[i].lineSpacing*=1.15;
	}

	menu.background=kFileErr;
	menu.RenderCallback=InterfaceRenderReplay;

	InterfaceMenuZoomAnimation(&menu,-1,true);	
	int exit=false;
	while(!exit)
		switch(menu.initialSelection=InterfaceGetUserMenuSelection(&menu))
		{
			case kInterfaceMultiplayerJoinGameTracker:
				InterfaceFindGame();
				break;
				
			case kInterfaceMultiplayerJoinGameManually:
				ManuallyJoinGame();
				break;

			case kInterfaceMultiplayerHostGame:
				InterfaceSetupHost();
				break;
				
			case kInterfaceMultiplayerReturn:
			case kInterfaceMenuEsc:
				exit=true;
				break;
		}
	InterfaceDisposeMenu(&menu);
}*/
